#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.conf import settings

from .common import BaseBackendTestCase
from .common import ONE_QUEUE_TWO_WORKERS

from ...experiments.models import Experiment
from ...experiments.models import Block
from ...experiments.models import CachedFile

from ..models import Job
from ..models import JobSplit
from ..models import Result
from ..models import Worker
from ..models import Queue

from ..helpers import schedule_experiment
from ..helpers import cancel_experiment
from ..helpers import split_new_jobs
from ..helpers import process_newly_cancelled_experiments
from ..helpers import is_cache_complete
from ..helpers import assign_splits_to_workers
from ..helpers import get_configuration_for_split
from ..helpers import on_split_started
from ..helpers import on_split_done
from ..helpers import on_split_fail
from ..helpers import on_split_cancelled

from ..utils import setup_backend

import beat.core.hash

import os
import sys


#----------------------------------------------------------


class ScheduleExperimentTest(BaseBackendTestCase):

    def check_pending_block_of_pending_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        experiment = block.experiment

        self.assertEqual(block.status, Block.PENDING)
        self.assertTrue(block.creation_date is not None)
        self.assertTrue(block.start_date is None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(experiment.status, Experiment.PENDING)


    def check_pending_block_of_scheduled_experiment(self, block, runnable=True):
        block.refresh_from_db()
        block.job.refresh_from_db()
        block.experiment.refresh_from_db()

        job = block.job
        experiment = block.experiment

        self.assertEqual(block.status, Block.PENDING)
        self.assertTrue(block.creation_date is not None)
        self.assertTrue(block.start_date is None)
        self.assertTrue(block.end_date is None)

        if runnable:
            self.assertTrue(job.runnable_date is not None)
        else:
            self.assertTrue(job.runnable_date is None)

        self.assertTrue(job.start_date is None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(experiment.status, Experiment.SCHEDULED)


    def check_done_block_of_scheduled_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        experiment = block.experiment

        self.assertEqual(block.status, Block.DONE)
        self.assertTrue(block.creation_date is not None)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(experiment.status, Experiment.SCHEDULED)


    def check_done_block_of_done_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        experiment = block.experiment

        self.assertEqual(block.status, Block.DONE)
        self.assertTrue(block.creation_date is not None)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(experiment.status, Experiment.DONE)


    def test_success(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.check_pending_block_of_pending_experiment(b0)
        self.check_pending_block_of_pending_experiment(b1)

        self.assertEqual(Job.objects.count(), 0)

        schedule_experiment(xp)

        self.check_pending_block_of_scheduled_experiment(b0)
        self.check_pending_block_of_scheduled_experiment(b1, runnable=False)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_first_block_in_cache(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            cache_status={
                'echo': CachedFile.CACHED,
            }
        )

        schedule_experiment(xp)

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.check_done_block_of_scheduled_experiment(b0)
        self.check_pending_block_of_scheduled_experiment(b1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_all_blocks_in_cache(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            cache_status={
                'echo': CachedFile.CACHED,
                'analysis': CachedFile.CACHED,
            }
        )

        schedule_experiment(xp)

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.check_done_block_of_done_experiment(b0)
        self.check_done_block_of_done_experiment(b1)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_already_scheduled(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        schedule_experiment(xp)
        xp.refresh_from_db()

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.check_pending_block_of_scheduled_experiment(b0)
        self.check_pending_block_of_scheduled_experiment(b1, runnable=False)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_running_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            experiment_status=Experiment.RUNNING,
            block_status={
                'echo': Block.PROCESSING,
            },
            cache_status={
                'echo': CachedFile.PROCESSING,
            }
        )

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.PROCESSING)
        self.assertEqual(b1.status, Block.PENDING)
        self.assertEqual(xp.status, Experiment.RUNNING)

        # schedule_experiment() didn't do anything
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_done_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            experiment_status=Experiment.DONE,
            block_status={
                'echo': Block.DONE,
                'analysis': Block.DONE,
            },
            cache_status={
                'echo': CachedFile.CACHED,
                'analysis': CachedFile.CACHED,
            }
        )

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.DONE)
        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(xp.status, Experiment.DONE)

        # schedule_experiment() didn't do anything
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_failed_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            experiment_status=Experiment.FAILED,
            block_status={
                'echo': Block.FAILED,
            },
        )

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.FAILED)
        self.assertEqual(b1.status, Block.PENDING)
        self.assertEqual(xp.status, Experiment.FAILED)

        # schedule_experiment() didn't do anything
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_cancelling_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            experiment_status=Experiment.CANCELLING,
            block_status={
                'echo': Block.CANCELLED,
            },
        )

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.CANCELLED)
        self.assertEqual(b1.status, Block.PENDING)
        self.assertEqual(xp.status, Experiment.CANCELLING)

        # schedule_experiment() didn't do anything
        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_cancelled_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.set_experiment_state(
            xp,
            block_status={
                'echo': Block.CANCELLED,
                'analysis': Block.CANCELLED,
            },
        )

        schedule_experiment(xp)
        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.PENDING)
        self.assertEqual(b1.status, Block.PENDING)
        self.assertEqual(xp.status, Experiment.SCHEDULED)

        # schedule_experiment() didn't do anything
        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)


    def test_two_different_experiments(self):
        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_add'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]
        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        self.check_pending_block_of_scheduled_experiment(b1_0)
        self.check_pending_block_of_scheduled_experiment(b1_1, runnable=False)

        self.check_pending_block_of_scheduled_experiment(b2_0)
        self.check_pending_block_of_scheduled_experiment(b2_1, runnable=False)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.assertNotEqual(b1_0.job.key, b2_0.job.key)


    def test_two_similar_experiments(self):
        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_split_2'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]
        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        self.check_pending_block_of_scheduled_experiment(b1_0)
        self.check_pending_block_of_scheduled_experiment(b1_1, runnable=False)

        self.check_pending_block_of_scheduled_experiment(b2_0)
        self.check_pending_block_of_scheduled_experiment(b2_1, runnable=False)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.assertEqual(b1_0.job.key, b2_0.job.key)


#----------------------------------------------------------


class IsCacheCompleteTest(BaseBackendTestCase):

    def test_one_split(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        self.assertTrue(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 10))


    def test_two_splits(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
        ])

        self.assertTrue(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 20))


    def test_three_splits(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24],
        ])

        self.assertTrue(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_one_split_with_ranges(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [(0, 2), (3, 5), (6, 8)],
        ])

        self.assertTrue(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 3))


    def test_two_splits_with_ranges(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [(0, 2), (3, 5), (6, 8)],
            [(9, 11), (12, 20)],
        ])

        self.assertTrue(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 5))


    def test_missing_indexes(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.index')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_missing_indexes_checksum(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.index.checksum')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_missing_data(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.data')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_missing_data_checksum(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.data.checksum')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_missing_indexes_and_data(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.index')))
        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.index.checksum')))
        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.data')))
        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.0.9.data.checksum')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_splitted_missing_indexes(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.10.19.index')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_splitted_missing_indexes_checksum(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.10.19.index.checksum')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_splitted_missing_data(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.10.19.data')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_splitted_missing_data_checksum(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24],
        ])

        os.remove(os.path.join(settings.CACHE_ROOT, beat.core.hash.toPath(hash, suffix='.10.19.data.checksum')))

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


    def test_splitted_missing_indexes_and_data(self):
        hash = '0123456789ABCDEF'

        self.generate_cached_files(hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            [20, 21, 22, 23, 24],
        ])

        self.assertFalse(is_cache_complete(beat.core.hash.toPath(hash, suffix=''), 25))


#----------------------------------------------------------


class SplitNewJobsTest(BaseBackendTestCase):

    def check_split(self, split, split_index=None, start_index=None, end_index=None):
        self.assertTrue(split.worker is None)
        self.assertEqual(split.split_index, split_index)
        self.assertEqual(split.start_index, start_index)
        self.assertEqual(split.end_index, end_index)
        self.assertEqual(split.status, JobSplit.QUEUED)
        self.assertTrue(split.result is None)
        self.assertTrue(split.start_date is None)
        self.assertTrue(split.end_date is None)


    def test_one_experiment_one_slot(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 1)

        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]
        split = b0.job.splits.all()[0]

        self.check_split(split, split_index=0)


    def test_one_experiment_two_slots(self):
        fullname = 'user/user/single/1/single_split_2'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        b0 = xp.blocks.all()[0]

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 2)

        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]

        self.assertEqual(b0.job.splits.count(), 2)

        split1 = b0.job.splits.all()[0]
        split2 = b0.job.splits.all()[1]

        self.check_split(split1, split_index=0, start_index=0, end_index=4)
        self.check_split(split2, split_index=1, start_index=5, end_index=9)


    def test_two_different_experiments(self):
        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_add'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 0)

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        b0 = xp1.blocks.all()[0]
        split = b0.job.splits.all()[0]

        self.check_split(split, split_index=0)

        b0 = xp2.blocks.all()[0]
        split = b0.job.splits.all()[0]

        self.check_split(split, split_index=0)


    def test_two_similar_experiments(self):
        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_split_2'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 0)

        b0 = xp1.blocks.all()[0]

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 3)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        b0 = xp1.blocks.all()[0]

        self.assertEqual(b0.job.splits.count(), 1)

        split = b0.job.splits.all()[0]

        self.check_split(split, split_index=0)

        b0 = xp2.blocks.all()[0]

        self.assertEqual(b0.job.splits.count(), 2)

        split1 = b0.job.splits.all()[0]
        split2 = b0.job.splits.all()[1]

        self.check_split(split1, split_index=0, start_index=0, end_index=4)
        self.check_split(split2, split_index=1, start_index=5, end_index=9)


    def test_one_experiment_two_uneven_slots(self):
        fullname = 'user/user/duo/1/split_2'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        b0 = xp.blocks.all()[0]

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 2)

        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]

        self.assertEqual(b0.job.splits.count(), 2)

        split1 = b0.job.splits.all()[0]
        split2 = b0.job.splits.all()[1]

        print split2.start_index
        print split2.end_index

        self.check_split(split1, split_index=0, start_index=0, end_index=5)
        self.check_split(split2, split_index=1, start_index=6, end_index=8)


    def test_one_experiment_too_much_slots(self):
        fullname = 'user/user/single/1/single_split_10'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        b0 = xp.blocks.all()[0]

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        split_new_jobs()

        self.assertEqual(JobSplit.objects.count(), 5)

        xp.refresh_from_db()

        b0 = xp.blocks.all()[0]

        self.assertEqual(b0.job.splits.count(), 5)

        split1 = b0.job.splits.all()[0]
        split2 = b0.job.splits.all()[1]
        split3 = b0.job.splits.all()[2]
        split4 = b0.job.splits.all()[3]
        split5 = b0.job.splits.all()[4]

        self.check_split(split1, split_index=0, start_index=0, end_index=0)
        self.check_split(split2, split_index=1, start_index=1, end_index=1)
        self.check_split(split3, split_index=2, start_index=2, end_index=2)
        self.check_split(split4, split_index=3, start_index=3, end_index=3)
        self.check_split(split5, split_index=4, start_index=4, end_index=4)


    def test_similar_experiment_after_assignation(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=True)

        fullname = 'user/user/single/1/single'

        xp1 = Experiment.objects.get(name=fullname.split('/')[-1])
        xp2 = xp1.fork(name='single_fork')

        self.prepare_databases(xp1.declaration)

        schedule_experiment(xp1)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 1)
        self.assertEqual(len(assigned_splits), 1)

        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 1)

        split = assigned_splits[0]

        self.assertTrue(split.worker is not None)
        self.assertFalse(split.job.mirror)

        mirror_job = Job.objects.get(block__experiment=xp2, runnable_date__isnull=False)

        self.assertEqual(split.job.key, mirror_job.key)
        self.assertEqual(mirror_job.splits.count(), 0)
        self.assertTrue(mirror_job.mirror)


    def test_similar_experiment_during_processing(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=True)

        fullname = 'user/user/single/1/single'

        xp1 = Experiment.objects.get(name=fullname.split('/')[-1])
        xp2 = xp1.fork(name='single_fork')

        self.prepare_databases(xp1.declaration)

        schedule_experiment(xp1)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 1)
        self.assertEqual(len(assigned_splits), 1)

        split = assigned_splits[0]

        split.status = JobSplit.PROCESSING
        split.save()

        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 1)

        split = assigned_splits[0]

        self.assertTrue(split.worker is not None)
        self.assertFalse(split.job.mirror)

        mirror_job = Job.objects.get(block__experiment=xp2, runnable_date__isnull=False)

        self.assertEqual(split.job.key, mirror_job.key)
        self.assertEqual(mirror_job.splits.count(), 0)
        self.assertTrue(mirror_job.mirror)


    def test_similar_experiment_after_completion_of_first_block(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.set_experiment_state(
            xp,
            cache_status={
                'echo': CachedFile.CACHED,
            }
        )

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.generate_cached_files(b0.outputs.all()[0].hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        ])

        split_new_jobs()

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 1)

        xp.refresh_from_db()
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.DONE)
        self.assertEqual(b1.status, Block.PENDING)

        split = b1.job.splits.all()[0]

        self.check_split(split, split_index=0)


    def test_similar_experiment_after_completion(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.set_experiment_state(
            xp,
            cache_status={
                'echo': CachedFile.CACHED,
                'analysis': CachedFile.CACHED,
            }
        )

        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.generate_cached_files(b0.outputs.all()[0].hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        ])

        self.generate_cached_files(b1.outputs.all()[0].hash, [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        ])

        split_new_jobs()

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        xp.refresh_from_db()
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(b0.status, Block.DONE)
        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(xp.status, Experiment.DONE)


#----------------------------------------------------------


class AssignSplitsToWorkersTest(BaseBackendTestCase):

    def test_one_experiment_one_slot(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.filter(name='node1').update(active=True)

        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 1)
        self.assertEqual(len(assigned_splits), 1)

        split = assigned_splits[0]

        self.assertTrue(split.worker is not None)
        self.assertFalse(split.job.mirror)


    def test_two_different_experiments_one_slot(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.filter(name='node1').update(active=True)

        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_add'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(len(assigned_splits), 1)

        split = assigned_splits[0]

        self.assertTrue(split.worker is not None)
        self.assertFalse(split.job.mirror)


    def test_two_different_experiments_two_slots(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=True)

        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_add'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(len(assigned_splits), 2)

        split1 = assigned_splits[0]

        self.assertTrue(split1.worker is not None)
        self.assertFalse(split1.job.mirror)

        split2 = assigned_splits[1]

        self.assertTrue(split2.worker is not None)
        self.assertFalse(split2.job.mirror)

        self.assertNotEqual(split1.job.key, split2.job.key)
        self.assertNotEqual(split1.worker, split2.worker)


    def test_two_different_experiments_two_slots_user_limitation(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=True)
        Queue.objects.filter(name='queue').update(max_slots_per_user=1)

        fullname1 = 'user/user/single/1/single'
        fullname2 = 'user/user/single/1/single_add'

        xp1 = Experiment.objects.get(name=fullname1.split('/')[-1])
        xp2 = Experiment.objects.get(name=fullname2.split('/')[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(len(assigned_splits), 1)

        split1 = assigned_splits[0]

        self.assertTrue(split1.worker is not None)
        self.assertFalse(split1.job.mirror)


    def test_two_similar_experiments_one_slot(self):
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.filter(name='node1').update(active=True)

        fullname = 'user/user/single/1/single'

        xp1 = Experiment.objects.get(name=fullname.split('/')[-1])
        xp2 = xp1.fork(name='single_fork')

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        split_new_jobs()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        assigned_splits = assign_splits_to_workers()

        self.assertEqual(JobSplit.objects.count(), 1)
        self.assertEqual(len(assigned_splits), 1)

        split = assigned_splits[0]

        self.assertTrue(split.worker is not None)
        self.assertFalse(split.job.mirror)

        mirror_job = Job.objects.get(mirror=True)

        self.assertEqual(split.job.key, mirror_job.key)
        self.assertEqual(mirror_job.splits.count(), 0)


#----------------------------------------------------------


class GetConfigurationForSplitTest(BaseBackendTestCase):

    def setUp(self):
        super(GetConfigurationForSplitTest, self).setUp()

        self.previous_DATASETS_UID = settings.DATASETS_UID
        self.previous_DATASETS_ROOT_PATH = settings.DATASETS_ROOT_PATH

        settings.DATASETS_UID = None
        settings.DATASETS_ROOT_PATH = None


    def tearDown(self):
        super(GetConfigurationForSplitTest, self).tearDown()

        settings.DATASETS_UID = self.previous_DATASETS_UID
        settings.DATASETS_ROOT_PATH = self.previous_DATASETS_ROOT_PATH


    def prepare_experiment(self, name):
        xp = Experiment.objects.get(name=name.split('/')[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)

        b0 = xp.blocks.all()[0]

        split_new_jobs()

        xp.refresh_from_db()

        return xp


    def test_one_split(self):
        xp = self.prepare_experiment('user/user/single/1/single')

        b0 = xp.blocks.all()[0]
        split = b0.job.splits.all()[0]

        configuration = get_configuration_for_split(split)

        self.assertTrue('datasets_uid' not in configuration)
        self.assertTrue('datasets_root_path' not in configuration)
        self.assertTrue('range' not in configuration)


    def test_two_splits(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')

        b0 = xp.blocks.all()[0]
        split = b0.job.splits.all()[0]

        configuration = get_configuration_for_split(split)

        self.assertTrue('datasets_uid' not in configuration)
        self.assertTrue('datasets_root_path' not in configuration)
        self.assertTrue('range' in configuration)

        self.assertEqual(configuration['range'][0], 0)
        self.assertEqual(configuration['range'][1], 4)

        split = b0.job.splits.all()[1]

        configuration = get_configuration_for_split(split)

        self.assertTrue('datasets_uid' not in configuration)
        self.assertTrue('datasets_root_path' not in configuration)
        self.assertTrue('range' in configuration)

        self.assertEqual(configuration['range'][0], 5)
        self.assertEqual(configuration['range'][1], 9)


    def test_dataset_uid(self):
        settings.DATASETS_UID = 12345

        xp = self.prepare_experiment('user/user/single/1/single')

        b0 = xp.blocks.all()[0]
        split = b0.job.splits.all()[0]

        configuration = get_configuration_for_split(split)

        self.assertTrue('datasets_uid' in configuration)
        self.assertTrue('datasets_root_path' not in configuration)
        self.assertTrue('range' not in configuration)

        self.assertEqual(configuration['datasets_uid'], settings.DATASETS_UID)


    def test_datasets_root_path(self):
        settings.DATASETS_ROOT_PATH = '/some/path/'

        xp = self.prepare_experiment('user/user/single/1/single')

        b0 = xp.blocks.all()[0]
        split = b0.job.splits.all()[0]

        configuration = get_configuration_for_split(split)

        self.assertTrue('datasets_uid' not in configuration)
        self.assertTrue('datasets_root_path' in configuration)
        self.assertTrue('range' not in configuration)

        self.assertEqual(configuration['datasets_root_path'], settings.DATASETS_ROOT_PATH)


#----------------------------------------------------------


class SplitHelpersBaseTest(BaseBackendTestCase):

    def prepare_experiment(self, name, generate_cache=True):
        xp = Experiment.objects.get(name=name.split('/')[-1])

        if generate_cache:
            self.prepare_databases(xp.declaration)

        schedule_experiment(xp)

        b0 = xp.blocks.all()[0]

        split_new_jobs()

        xp.refresh_from_db()

        return xp


    def check_cached_file_statistics(self, reference, block, factor=1.0, cached=True):
        cpu = reference['statistics']['cpu']
        memory = reference['statistics']['memory']
        data = reference['statistics']['data']

        for cached_file in block.outputs.all():
            self.assertEqual(cached_file.status, CachedFile.CACHED if cached else CachedFile.NOT_CACHED)
            self.assertEqual(cached_file.cpu_time, (cpu['user'] + cpu['system']) * factor)
            self.assertEqual(cached_file.max_memory, memory['rss'] * factor)
            self.assertEqual(cached_file.data_read_size, data['volume']['read'] * factor)
            self.assertEqual(cached_file.data_read_nb_blocks, data['blocks']['read'] * factor)
            self.assertEqual(cached_file.data_read_time, data['time']['read'] * factor)
            self.assertEqual(cached_file.data_written_size, data['volume']['write'] * factor)
            self.assertEqual(cached_file.data_written_nb_blocks, data['blocks']['write'] * factor)
            self.assertEqual(cached_file.data_written_time, data['time']['write'] * factor)


    def check_split_before_start(self, split, experiment_status=Experiment.SCHEDULED):
        self.assertTrue(split.worker is not None)
        self.assertEqual(split.status, JobSplit.QUEUED)
        self.assertTrue(split.start_date is None)
        self.assertTrue(split.end_date is None)

        self.check_job_before_start(split.job, experiment_status)


    def check_job_before_start(self, job, experiment_status=Experiment.SCHEDULED):
        self.assertTrue(job.start_date is None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(job.block.status, Block.PENDING)
        self.assertTrue(job.block.start_date is None)
        self.assertTrue(job.block.end_date is None)

        self.assertEqual(job.block.experiment.status, experiment_status)

        if experiment_status == Experiment.SCHEDULED:
            self.assertTrue(job.block.experiment.start_date is None)
            self.assertTrue(job.block.experiment.end_date is None)

        elif experiment_status == Experiment.RUNNING:
            self.assertTrue(job.block.experiment.start_date is not None)
            self.assertTrue(job.block.experiment.end_date is None)


    def check_processing_split(self, split):
        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.PROCESSING)
        self.assertTrue(split.start_date is not None)
        self.assertTrue(split.end_date is None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(split.start_date >= job.start_date)
        self.assertEqual(block.start_date, job.start_date)
        self.assertTrue(job.start_date >= xp.start_date)


    def check_queued_split_of_running_job(self, split):
        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.QUEUED)
        self.assertTrue(split.start_date is None)
        self.assertTrue(split.end_date is None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertEqual(xp.start_date, job.start_date)


    def check_done_split_of_running_job(self, split):
        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.COMPLETED)
        self.assertTrue(split.start_date is not None)
        self.assertTrue(split.end_date is not None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(split.start_date >= job.start_date)
        self.assertEqual(block.start_date, job.start_date)
        self.assertEqual(xp.start_date, job.start_date)


    def check_failed_split_of_running_job(self, split):
        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.FAILED)
        self.assertTrue(split.start_date is not None)
        self.assertTrue(split.end_date is not None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(split.end_date >= split.start_date)
        self.assertTrue(split.start_date >= job.start_date)
        self.assertTrue(job.start_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_cancelling_split_of_cancelling_experiment(
            self, split, experiment_status=Experiment.RUNNING):
        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.CANCELLING)
        self.assertTrue(split.start_date is not None)
        self.assertTrue(split.end_date is None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, experiment_status)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(split.start_date >= job.start_date)
        self.assertTrue(job.start_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_cancelled_split_of_cancelling_experiment(
        self, split, experiment_status=Experiment.RUNNING):

        split.refresh_from_db()
        split.job.refresh_from_db()
        split.job.block.refresh_from_db()
        split.job.block.experiment.refresh_from_db()

        job = split.job
        block = job.block
        xp = block.experiment

        self.assertEqual(split.status, JobSplit.CANCELLED)
        self.assertTrue(split.start_date is not None)
        self.assertTrue(split.end_date is not None)

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, experiment_status)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(split.end_date >= split.start_date)
        self.assertTrue(split.start_date >= job.start_date)
        self.assertTrue(job.start_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_processing_mirror_job(self, job):
        job.refresh_from_db()
        job.block.refresh_from_db()
        job.block.experiment.refresh_from_db()

        block = job.block
        xp = block.experiment

        self.assertTrue(job.start_date is not None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PROCESSING)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertEqual(block.start_date, job.start_date)
        self.assertEqual(xp.start_date, job.start_date)


    def check_queued_job_of_running_experiment(self, job):
        job.refresh_from_db()
        job.block.refresh_from_db()
        job.block.experiment.refresh_from_db()

        block = job.block
        xp = block.experiment

        self.assertTrue(job.start_date is None)
        self.assertTrue(job.end_date is None)

        self.assertEqual(block.status, Block.PENDING)
        self.assertTrue(block.start_date is None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)


    def check_done_block_of_running_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.DONE)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_pending_block_of_running_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.PENDING)
        self.assertTrue(block.start_date is None)
        self.assertTrue(block.end_date is None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)


    def check_done_block_of_complete_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.DONE)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.DONE)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is not None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)
        self.assertTrue(xp.end_date >= block.end_date)


    def check_failed_block_of_failed_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.FAILED)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.FAILED)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is not None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)
        self.assertTrue(xp.end_date >= block.end_date)


    def check_cancelled_block_of_failed_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.CANCELLED)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.FAILED)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is not None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)
        self.assertTrue(xp.end_date >= block.end_date)


    def check_failed_block_of_cancelling_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.FAILED)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.RUNNING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_cancelled_block_of_cancelling_experiment(
            self, block, experiment_status=Experiment.RUNNING):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.CANCELLED)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, experiment_status)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)


    def check_cancelled_block_of_cancelled_experiment(self, block):
        block.refresh_from_db()
        block.experiment.refresh_from_db()

        xp = block.experiment

        self.assertEqual(block.status, Block.CANCELLED)
        self.assertTrue(block.start_date is not None)
        self.assertTrue(block.end_date is not None)

        self.assertEqual(xp.status, Experiment.PENDING)
        self.assertTrue(xp.start_date is not None)
        self.assertTrue(xp.end_date is not None)

        self.assertTrue(block.end_date >= block.start_date)
        self.assertTrue(block.start_date >= xp.start_date)
        self.assertTrue(xp.end_date >= block.end_date)


#----------------------------------------------------------


class OnSplitStartedTest(SplitHelpersBaseTest):

    def test_one_split(self):
        xp = self.prepare_experiment('user/user/single/1/single')

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.check_split_before_start(split)

        on_split_started(split)

        self.check_processing_split(split)


    def test_two_splits(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')

        assigned_splits = assign_splits_to_workers()

        split0 = assigned_splits[0]
        split1 = assigned_splits[1]

        self.check_split_before_start(split0)
        self.check_split_before_start(split1)

        on_split_started(split0)

        self.check_processing_split(split0)
        self.check_queued_split_of_running_job(split1)

        on_split_started(split1)

        self.check_processing_split(split0)
        self.check_processing_split(split1)

        split0.refresh_from_db()
        split1.refresh_from_db()

        self.assertTrue(split1.start_date > split0.start_date)


    def test_mirror_jobs_one_split(self):
        xp1 = self.prepare_experiment('user/user/single/1/single')
        xp2 = self.prepare_experiment('user/user/single/1/single_split_2',
                                      generate_cache=False)

        assigned_splits = assign_splits_to_workers()

        b1 = xp1.blocks.all()[0]
        b2 = xp2.blocks.all()[0]
        split = assigned_splits[0]

        self.assertEqual(split.job, b1.job)

        self.check_split_before_start(split)
        self.check_job_before_start(b2.job)

        on_split_started(split)

        self.check_processing_split(split)
        self.check_processing_mirror_job(b2.job)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        self.assertEqual(xp2.start_date, xp1.start_date)


    def test_mirror_jobs_two_splits(self):
        xp1 = self.prepare_experiment('user/user/single/1/single_split_2')
        xp2 = self.prepare_experiment('user/user/single/1/single',
                                      generate_cache=False)

        assigned_splits = assign_splits_to_workers()

        b1 = xp1.blocks.all()[0]
        b2 = xp2.blocks.all()[0]
        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        self.assertEqual(split1.job, b1.job)
        self.assertEqual(split2.job, b1.job)

        self.check_split_before_start(split1)
        self.check_split_before_start(split2)
        self.check_job_before_start(b2.job)

        on_split_started(split1)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        self.check_processing_split(split1)
        self.check_queued_split_of_running_job(split2)
        self.check_processing_mirror_job(b2.job)

        self.assertEqual(xp2.start_date, xp1.start_date)

        on_split_started(split2)

        self.check_processing_split(split1)
        self.check_processing_split(split2)
        self.check_processing_mirror_job(b2.job)

        xp1.refresh_from_db()
        xp2.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertTrue(split2.start_date > split1.start_date)
        self.assertEqual(xp2.start_date, xp1.start_date)


#----------------------------------------------------------


class OnSplitDoneTest(SplitHelpersBaseTest):

    RESULT = dict(
        status = 0,
        stdout = '',
        stderr = '',
        user_error = '',
        statistics = dict(
            data = dict(
                volume = dict(
                    read = 100,
                    write = 50
                ),
                blocks = dict(
                    read = 10,
                    write = 5
                ),
                time = dict(
                    read = 20,
                    write = 15
                ),
                network = dict(
                    wait_time = 4
                ),
                files = [],
            ),
            cpu = dict(
                user = 30.0,
                system = 40.0,
                total = 80.0,
                percent = 0.0,
                processors = 1,
            ),
            memory = dict(
                rss = 1024,
                limit = 2048,
                percent = 50.0,
            ),
        )
    )


    def check_cached_file_statistics(self, block, factor=1.0):
        super(OnSplitDoneTest, self).check_cached_file_statistics(
            OnSplitDoneTest.RESULT, block, factor=factor
        )


    def test_one_split_per_block(self):
        xp = self.prepare_experiment('user/user/single/1/single')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(Job.objects.count(), 2)

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.check_split_before_start(split)

        self.assertEqual(split.job.block, b0)

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 1)

        self.check_done_block_of_running_experiment(b0)
        self.check_pending_block_of_running_experiment(b1)

        self.check_cached_file_statistics(b0)

        # Block 1
        split_new_jobs()
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.check_split_before_start(split, experiment_status=Experiment.RUNNING)

        self.assertEqual(split.job.block, b1)

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 0)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.check_done_block_of_complete_experiment(b0)
        self.check_done_block_of_complete_experiment(b1)

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)

        self.check_cached_file_statistics(b1)


    def test_two_splits_for_first_block_sequential(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(Job.objects.count(), 2)

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        self.check_split_before_start(split1)
        self.check_split_before_start(split2)

        self.assertEqual(split1.job.block, b0)
        self.assertEqual(split2.job.block, b0)

        # Block 0 - Split 0
        on_split_started(split1)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.check_processing_split(split1)

        on_split_done(split1, OnSplitDoneTest.RESULT)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(Job.objects.count(), 2)

        self.check_done_split_of_running_job(split1)
        self.check_queued_split_of_running_job(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        self.assertEqual(xp.start_date, b0.start_date)

        # Block 0 - Split 1
        on_split_started(split2)

        xp.refresh_from_db()
        split2.refresh_from_db()

        self.check_processing_split(split2)

        on_split_done(split2, OnSplitDoneTest.RESULT)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Job.objects.count(), 1)

        self.check_done_block_of_running_experiment(b0)
        self.check_queued_job_of_running_experiment(b1.job)

        self.assertEqual(xp.start_date, b0.start_date)

        self.check_cached_file_statistics(b0, factor=2)

        # Block 1
        split_new_jobs()
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.check_split_before_start(split, experiment_status=Experiment.RUNNING)

        self.assertEqual(split.job.block, b1)

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 0)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.check_done_block_of_complete_experiment(b0)
        self.check_done_block_of_complete_experiment(b1)

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)

        self.check_cached_file_statistics(b1)


    def test_two_splits_for_first_block_in_parallel(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(Job.objects.count(), 2)

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        self.check_split_before_start(split1)
        self.check_split_before_start(split2)

        self.assertEqual(split1.job.block, b0)
        self.assertEqual(split2.job.block, b0)

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.check_processing_split(split1)
        self.check_processing_split(split2)

        on_split_done(split1, OnSplitDoneTest.RESULT)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(JobSplit.objects.count(), 2)
        self.assertEqual(Job.objects.count(), 2)

        self.check_done_split_of_running_job(split1)
        self.check_processing_split(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        self.assertEqual(xp.start_date, b0.start_date)

        on_split_done(split2, OnSplitDoneTest.RESULT)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(JobSplit.objects.count(), 0)
        self.assertEqual(Job.objects.count(), 1)

        self.check_done_block_of_running_experiment(b0)
        self.check_queued_job_of_running_experiment(b1.job)

        self.assertEqual(xp.start_date, b0.start_date)

        self.check_cached_file_statistics(b0, factor=2)

        # Block 1
        split_new_jobs()
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.check_split_before_start(split, experiment_status=Experiment.RUNNING)

        self.assertEqual(split.job.block, b1)

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 0)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.check_done_block_of_complete_experiment(b0)
        self.check_done_block_of_complete_experiment(b1)

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)

        self.check_cached_file_statistics(b1)


    def test_mirror_experiments_one_split_per_block(self):
        xp1 = self.prepare_experiment('user/user/single/1/single')
        xp2 = self.prepare_experiment('user/user/single/1/single_split_2',
                                      generate_cache=False)
        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]
        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        self.assertEqual(Job.objects.count(), 4)

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.assertEqual(split.job.block, b1_0)

        self.check_split_before_start(split)
        self.check_job_before_start(b2_0.job)

        on_split_started(split)

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 2)

        self.check_done_block_of_running_experiment(b1_0)
        self.check_pending_block_of_running_experiment(b1_1)

        self.check_done_block_of_running_experiment(b2_0)
        self.check_pending_block_of_running_experiment(b2_1)

        self.check_cached_file_statistics(b1_0)

        # Block 1
        split_new_jobs()
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.assertEqual(split.job.block, b1_1)

        self.check_split_before_start(split, experiment_status=Experiment.RUNNING)
        self.check_job_before_start(b2_1.job, experiment_status=Experiment.RUNNING)

        on_split_started(split)

        self.check_processing_split(split)

        on_split_done(split, OnSplitDoneTest.RESULT)

        self.assertEqual(Job.objects.count(), 0)

        xp1.refresh_from_db()
        xp2.refresh_from_db()
        b1_0.refresh_from_db()
        b1_1.refresh_from_db()
        b2_0.refresh_from_db()
        b2_1.refresh_from_db()

        self.check_done_block_of_complete_experiment(b1_0)
        self.check_done_block_of_complete_experiment(b1_1)
        self.check_done_block_of_complete_experiment(b2_0)
        self.check_done_block_of_complete_experiment(b2_1)

        self.assertEqual(xp1.start_date, b1_0.start_date)
        self.assertEqual(xp1.end_date, b1_1.end_date)

        self.assertEqual(xp2.start_date, b2_0.start_date)
        self.assertEqual(xp2.end_date, b2_1.end_date)

        self.assertEqual(xp2.start_date, xp1.start_date)
        self.assertEqual(xp2.end_date, xp1.end_date)

        self.check_cached_file_statistics(b1_1)


#----------------------------------------------------------


class OnSplitFailTest(SplitHelpersBaseTest):

    RESULT = dict(
        status = 1,
        stdout = '',
        stderr = '',
        user_error = 'Some error',
        statistics = dict(
            data = dict(
                volume = dict(
                    read = 100,
                    write = 50
                ),
                blocks = dict(
                    read = 10,
                    write = 5
                ),
                time = dict(
                    read = 20,
                    write = 15
                ),
                network = dict(
                    wait_time = 4
                ),
                files = [],
            ),
            cpu = dict(
                user = 30.0,
                system = 40.0,
                total = 80.0,
                percent = 0.0,
                processors = 1,
            ),
            memory = dict(
                rss = 1024,
                limit = 2048,
                percent = 50.0,
            ),
        )
    )


    def check_cached_file_statistics(self, block, factor=1.0, cached=True):
        super(OnSplitFailTest, self).check_cached_file_statistics(
            OnSplitFailTest.RESULT, block, factor=factor, cached=cached
        )


    def test_system_error(self):
        xp = self.prepare_experiment('user/user/single/1/single_error')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)

        splits_to_cancel = on_split_fail(split, "Some error")

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, cached=False, factor=0)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_one_split_per_block(self):
        xp = self.prepare_experiment('user/user/single/1/single_error')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)

        splits_to_cancel = on_split_fail(split, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, cached=False)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_only_one_running(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        xp.refresh_from_db()
        split.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, cached=False)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_running_splits(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)
        self.assertEqual(splits_to_cancel[0], split2)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_failed_split_of_running_job(split1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_with_one_failure_and_one_success(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)
        self.assertEqual(splits_to_cancel[0], split2)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_failed_split_of_running_job(split1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        # Surprise: split2 sucessfully finished before we could cancel it
        on_split_done(split2, OnSplitFailTest.RESULT)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, factor=2, cached=False)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_with_one_success_and_one_failure(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        on_split_done(split1, OnSplitFailTest.RESULT)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_done_split_of_running_job(split1)

        splits_to_cancel = on_split_fail(split2, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, factor=2, cached=False)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_both_failing(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)
        self.assertEqual(splits_to_cancel[0], split2)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_failed_split_of_running_job(split1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        # Surprise: split2 failed before we could cancel it
        splits_to_cancel = on_split_fail(split2, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        self.check_cached_file_statistics(b0, factor=2, cached=False)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_mirror_experiments_one_split_per_block(self):
        xp1 = self.prepare_experiment('user/user/single/1/single_error')
        xp2 = self.prepare_experiment('user/user/single/1/single_error_split_2',
                                      generate_cache=False)
        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]
        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 1)

        splits_to_cancel = on_split_fail(split, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b1_0)
        self.check_cancelled_block_of_failed_experiment(b1_1)

        self.check_failed_block_of_failed_experiment(b2_0)
        self.check_cancelled_block_of_failed_experiment(b2_1)

        self.check_cached_file_statistics(b1_0, cached=False)

        xp1.refresh_from_db()
        xp2.refresh_from_db()
        b1_0.refresh_from_db()
        b1_1.refresh_from_db()
        b2_0.refresh_from_db()
        b2_1.refresh_from_db()

        self.assertEqual(xp1.start_date, b1_0.start_date)
        self.assertEqual(xp1.end_date, b1_1.end_date)

        self.assertEqual(xp2.start_date, b2_0.start_date)
        self.assertEqual(xp2.end_date, b2_1.end_date)

        self.assertEqual(xp2.start_date, xp1.start_date)
        self.assertTrue(xp2.end_date >= xp1.end_date)


    def test_two_blocks_in_parallel_one_split_per_block_only_one_running(self):
        xp = self.prepare_experiment('user/user/triangle/1/triangle')
        b1 = xp.blocks.get(name='echo1')
        b2 = xp.blocks.get(name='echo2')
        b3 = xp.blocks.get(name='echo3')
        b4 = xp.blocks.get(name='analysis')

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)

        xp.refresh_from_db()
        split1.refresh_from_db()

        self.assertEqual(Job.objects.count(), 4)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b1)
        self.check_cancelled_block_of_failed_experiment(b2)
        self.check_cancelled_block_of_failed_experiment(b3)
        self.check_cancelled_block_of_failed_experiment(b4)

        self.check_cached_file_statistics(b1, cached=False)

        xp.refresh_from_db()
        b1.refresh_from_db()
        b2.refresh_from_db()
        b3.refresh_from_db()
        b4.refresh_from_db()

        self.assertEqual(xp.start_date, b1.start_date)
        self.assertEqual(b2.start_date, b2.end_date)
        self.assertEqual(b3.start_date, b3.end_date)
        self.assertEqual(b4.start_date, b4.end_date)
        self.assertEqual(xp.end_date, b4.end_date)


    def test_two_blocks_in_parallel_one_split_per_block_both_running(self):
        xp = self.prepare_experiment('user/user/triangle/1/triangle')
        b1 = xp.blocks.get(name='echo1')
        b2 = xp.blocks.get(name='echo2')
        b3 = xp.blocks.get(name='echo3')
        b4 = xp.blocks.get(name='analysis')

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 4)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)

        split2.refresh_from_db()

        self.assertEqual(splits_to_cancel[0], split2)

        self.check_failed_block_of_cancelling_experiment(b1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_cancelled_block_of_cancelling_experiment(b3)
        self.check_cancelled_block_of_cancelling_experiment(b4)

        self.check_cached_file_statistics(b1, cached=False)

        xp.refresh_from_db()
        b1.refresh_from_db()
        b2.refresh_from_db()
        b3.refresh_from_db()
        b4.refresh_from_db()

        self.assertEqual(xp.start_date, b1.start_date)
        self.assertEqual(b3.start_date, b3.end_date)
        self.assertEqual(b4.start_date, b4.end_date)


#----------------------------------------------------------


class CancelExperimentTest(SplitHelpersBaseTest):

    def test_pending_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        cancel_experiment(xp)

        # No change
        self.assertEqual(xp.status, Experiment.PENDING)


    def test_scheduled_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        schedule_experiment(xp)

        cancel_experiment(xp)

        xp.refresh_from_db()

        self.assertEqual(xp.status, Experiment.CANCELLING)


    def test_running_experiment(self):
        xp = self.prepare_experiment('user/user/single/1/single')

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        cancel_experiment(xp)

        xp.refresh_from_db()

        self.assertEqual(xp.status, Experiment.CANCELLING)


    def test_done_experiment(self):
        xp = self.prepare_experiment('user/user/single/1/single')

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)
        on_split_done(split, OnSplitDoneTest.RESULT)

        split_new_jobs()
        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)
        on_split_done(split, OnSplitDoneTest.RESULT)

        cancel_experiment(xp)

        xp.refresh_from_db()

        self.assertEqual(xp.status, Experiment.DONE)


    def test_failed_experiment(self):
        xp = self.prepare_experiment('user/user/single/1/single')

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)
        on_split_fail(split, OnSplitFailTest.RESULT)

        cancel_experiment(xp)

        xp.refresh_from_db()

        self.assertEqual(xp.status, Experiment.FAILED)


    def test_cancelling_experiment(self):
        fullname = 'user/user/single/1/single'

        xp = Experiment.objects.get(name=fullname.split('/')[-1])

        schedule_experiment(xp)

        cancel_experiment(xp)

        cancel_experiment(xp)

        xp.refresh_from_db()

        self.assertEqual(xp.status, Experiment.CANCELLING)


#----------------------------------------------------------


class ProcessNewlyCancelledExperimentsTest(SplitHelpersBaseTest):

    def test_no_split_running(self):
        xp = self.prepare_experiment('user/user/single/1/single')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        cancel_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 0)

        self.check_cancelled_block_of_cancelled_experiment(b0)
        self.check_cancelled_block_of_cancelled_experiment(b1)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(b1.start_date, b1.end_date)
        self.assertEqual(xp.end_date, b1.end_date)


    def test_one_split_running(self):
        xp = self.prepare_experiment('user/user/single/1/single')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        cancel_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)

        self.check_cancelling_split_of_cancelling_experiment(split, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b1, experiment_status=Experiment.CANCELLING)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_one_split_of_two_in_a_block_running(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)

        cancel_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_cancelling_split_of_cancelling_experiment(split1, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_split_of_cancelling_experiment(split2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b1, experiment_status=Experiment.CANCELLING)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_of_same_block_running(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        cancel_experiment(xp)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 2)

        self.assertEqual(Job.objects.count(), 1)

        self.check_cancelling_split_of_cancelling_experiment(split1, experiment_status=Experiment.CANCELLING)
        self.check_cancelling_split_of_cancelling_experiment(split2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b1, experiment_status=Experiment.CANCELLING)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_two_splits_of_different_blocks_running(self):
        xp = self.prepare_experiment('user/user/triangle/1/triangle')
        b0 = xp.blocks.get(name='echo1')
        b1 = xp.blocks.get(name='echo2')
        b2 = xp.blocks.get(name='echo3')
        b3 = xp.blocks.get(name='analysis')

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        cancel_experiment(xp)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 2)

        self.assertEqual(Job.objects.count(), 2)

        self.check_cancelling_split_of_cancelling_experiment(split1, experiment_status=Experiment.CANCELLING)
        self.check_cancelling_split_of_cancelling_experiment(split2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b3, experiment_status=Experiment.CANCELLING)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()
        b2.refresh_from_db()
        b3.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b2.start_date, b2.end_date)
        self.assertEqual(b3.start_date, b3.end_date)


    def test_mirror_block(self):
        xp1 = self.prepare_experiment('user/user/single/1/single')
        xp2 = self.prepare_experiment('user/user/single/1/single_split_2',
                                      generate_cache=False)
        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]

        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.assertEqual(split.job.block, b1_0)

        on_split_started(split)

        cancel_experiment(xp2)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 1)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        self.check_cancelled_block_of_cancelled_experiment(b2_0)
        self.check_cancelled_block_of_cancelled_experiment(b2_1)

        self.check_processing_split(split)
        self.check_queued_job_of_running_experiment(b1_1.job)

        xp2.refresh_from_db()
        b2_0.refresh_from_db()
        b2_1.refresh_from_db()

        self.assertEqual(xp2.start_date, b2_0.start_date)
        self.assertEqual(b2_1.start_date, b2_1.end_date)


    def test_mirrored_block(self):
        xp1 = self.prepare_experiment('user/user/single/1/single')
        xp2 = self.prepare_experiment('user/user/single/1/single_split_2',
                                      generate_cache=False)
        b1_0 = xp1.blocks.all()[0]
        b1_1 = xp1.blocks.all()[1]

        b2_0 = xp2.blocks.all()[0]
        b2_1 = xp2.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        self.assertEqual(split.job.block, b1_0)

        on_split_started(split)

        cancel_experiment(xp1)

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 1)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 0)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 1)

        split.refresh_from_db()
        self.assertEqual(split.job.block, b2_0)

        self.check_cancelled_block_of_cancelled_experiment(b1_0)
        self.check_cancelled_block_of_cancelled_experiment(b1_1)

        self.check_processing_split(split)
        self.check_queued_job_of_running_experiment(b2_1.job)

        xp1.refresh_from_db()
        b1_0.refresh_from_db()
        b1_1.refresh_from_db()

        self.assertEqual(xp1.start_date, b1_0.start_date)
        self.assertEqual(b1_1.start_date, b1_1.end_date)


#----------------------------------------------------------


class OnSplitCancelledTest(SplitHelpersBaseTest):

    def check_cached_file_statistics(self, block, factor=1.0, cached=True):
        super(OnSplitCancelledTest, self).check_cached_file_statistics(
            OnSplitFailTest.RESULT, block, factor=factor, cached=cached
        )


    def test_failure_in_sibling_split(self):
        xp = self.prepare_experiment('user/user/single/1/single_error_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split1.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)
        self.assertEqual(splits_to_cancel[0], split2)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        self.check_failed_split_of_running_job(split1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_queued_job_of_running_experiment(b1.job)

        on_split_cancelled(split2)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b0)
        self.check_cancelled_block_of_failed_experiment(b1)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(xp.end_date, b1.end_date)
        self.assertEqual(b1.start_date, b1.end_date)


    def test_failure_in_parallel_block(self):
        xp = self.prepare_experiment('user/user/triangle/1/triangle')
        b1 = xp.blocks.get(name='echo1')
        b2 = xp.blocks.get(name='echo2')
        b3 = xp.blocks.get(name='echo3')
        b4 = xp.blocks.get(name='analysis')

        # Block 0
        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        xp.refresh_from_db()
        split2.refresh_from_db()

        self.assertEqual(Job.objects.count(), 4)
        self.assertEqual(JobSplit.objects.count(), 2)

        splits_to_cancel = on_split_fail(split1, OnSplitFailTest.RESULT)

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 1)

        split2.refresh_from_db()

        self.assertEqual(splits_to_cancel[0], split2)

        self.check_failed_block_of_cancelling_experiment(b1)
        self.check_cancelling_split_of_cancelling_experiment(split2)
        self.check_cancelled_block_of_cancelling_experiment(b3)
        self.check_cancelled_block_of_cancelling_experiment(b4)

        self.check_cached_file_statistics(b1, cached=False)

        on_split_cancelled(split2)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_failed_block_of_failed_experiment(b1)
        self.check_cancelled_block_of_failed_experiment(b2)
        self.check_cancelled_block_of_failed_experiment(b3)
        self.check_cancelled_block_of_failed_experiment(b4)

        xp.refresh_from_db()
        b1.refresh_from_db()
        b2.refresh_from_db()
        b3.refresh_from_db()
        b4.refresh_from_db()

        self.assertEqual(xp.start_date, b1.start_date)
        self.assertEqual(b3.start_date, b3.end_date)
        self.assertEqual(b4.start_date, b4.end_date)
        self.assertEqual(xp.end_date, b2.end_date)


    def test_cancelled_experiment_one_split_running(self):
        xp = self.prepare_experiment('user/user/single/1/single')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split = assigned_splits[0]

        on_split_started(split)

        cancel_experiment(xp)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 1)

        on_split_cancelled(splits_to_cancel[0])

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_cancelled_block_of_cancelled_experiment(b0)
        self.check_cancelled_block_of_cancelled_experiment(b1)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)
        self.assertEqual(xp.end_date, b0.end_date)


    def test_cancelled_experiment_one_split_of_two_in_a_block_running(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)

        cancel_experiment(xp)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 2)

        on_split_cancelled(splits_to_cancel[0])

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_cancelled_block_of_cancelled_experiment(b0)
        self.check_cancelled_block_of_cancelled_experiment(b1)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)
        self.assertEqual(xp.end_date, b0.end_date)


    def test_cancelled_experiment_two_splits_of_same_block_running(self):
        xp = self.prepare_experiment('user/user/single/1/single_split_2')
        b0 = xp.blocks.all()[0]
        b1 = xp.blocks.all()[1]

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        cancel_experiment(xp)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 2)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 2)

        on_split_cancelled(split1)

        self.check_cancelled_split_of_cancelling_experiment(split1, experiment_status=Experiment.CANCELLING)
        self.check_cancelling_split_of_cancelling_experiment(split2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b1, experiment_status=Experiment.CANCELLING)

        on_split_cancelled(split2)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_cancelled_block_of_cancelled_experiment(b0)
        self.check_cancelled_block_of_cancelled_experiment(b1)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b1.start_date, b1.end_date)
        self.assertEqual(xp.end_date, b0.end_date)


    def test_two_splits_of_different_blocks_running(self):
        xp = self.prepare_experiment('user/user/triangle/1/triangle')
        b0 = xp.blocks.get(name='echo1')
        b1 = xp.blocks.get(name='echo2')
        b2 = xp.blocks.get(name='echo3')
        b3 = xp.blocks.get(name='analysis')

        assigned_splits = assign_splits_to_workers()

        split1 = assigned_splits[0]
        split2 = assigned_splits[1]

        on_split_started(split1)
        on_split_started(split2)

        cancel_experiment(xp)

        splits_to_cancel = process_newly_cancelled_experiments()

        self.assertEqual(len(splits_to_cancel), 2)

        self.assertEqual(Job.objects.count(), 2)
        self.assertEqual(JobSplit.objects.count(), 2)

        on_split_cancelled(split1)

        self.assertEqual(Job.objects.count(), 1)
        self.assertEqual(JobSplit.objects.count(), 1)

        self.check_cancelled_block_of_cancelling_experiment(b0, experiment_status=Experiment.CANCELLING)
        self.check_cancelling_split_of_cancelling_experiment(split2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b2, experiment_status=Experiment.CANCELLING)
        self.check_cancelled_block_of_cancelling_experiment(b3, experiment_status=Experiment.CANCELLING)

        on_split_cancelled(split2)

        self.assertEqual(Job.objects.count(), 0)
        self.assertEqual(JobSplit.objects.count(), 0)

        self.check_cancelled_block_of_cancelled_experiment(b0)
        self.check_cancelled_block_of_cancelled_experiment(b1)
        self.check_cancelled_block_of_cancelled_experiment(b2)
        self.check_cancelled_block_of_cancelled_experiment(b3)

        xp.refresh_from_db()
        b0.refresh_from_db()
        b1.refresh_from_db()
        b2.refresh_from_db()
        b3.refresh_from_db()

        self.assertEqual(xp.start_date, b0.start_date)
        self.assertEqual(b2.start_date, b2.end_date)
        self.assertEqual(b3.start_date, b3.end_date)
        self.assertEqual(xp.end_date, b1.end_date)
